/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * aux_c: includes lexical analyzer for ZXSimulator, a ZX81 ROM emulator
 */


#include stdio_h
#include cmd_h


/* gets keyword from command line */
getkey (key)
   char key[];
{
   if (getsym(key) == NOERR)
      { if (strchr(delim,str[sp]) != NULL) if (!op || strchr(delim,str[op-1]) != NULL)
           return NOERR;
        sp = op; }

   return BADCHAR;
}


/* gets symbol from command line */
getsym (sym)
   char sym[];
{
   char ch;
   int i;

   while (str[sp] == ' ') { sp++; } op = sp;
   i = 0;
   while ((ch=sym[i])!='\0')
      { if (str[sp] != ch) { sp = op; return BADCHAR; } sp++; i++; }

   return NOERR;
}


/* displays formatted argument for print statement */
putformat ()
{
   long val_i;
   int  i, err, onLF;
   char ch, *val_s;

   err = NOERR;
   onLF = -2;   /* neither no-line-feed (-1) or maybe-line-feed (0) so print is ok */
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case 'A':
            if (getkey("AT") != NOERR) break;
            if (onLF >= 0 || getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            row = FIRSTROW + (labs((val_i+50L)/100L) << 3);
            if (getsym(",") != NOERR) return BADARG;
            if (getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            col = FIRSTCOL + (labs((val_i+50L)/100L) << 3);
            if (row > LASTROW || col > LASTCOL) return BADARG;
            onLF = 1;
            continue;
         case 'T':
            if (getkey("TAB") != NOERR) break;
            if (onLF >= 0 || getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            val_i = (val_i + 50L) / 100L;
            if (val_i < 0) return BADARG;
            if ((col >> 3) > val_i) row = row + 8;
            col = FIRSTCOL + ((val_i % COLLEN) << 3);
            onLF = 1;
            continue;
         case ',':
            if (col < MIDCOL)
               { col = MIDCOL; if (row > endrow) return NOMORE; }
            else
               { ZXdowncurs(); }
         case ';':
            sp++; onLF = -1;
            continue;
         case EOL:
            switch (onLF)
            {
               case -1:
                  break;
               case 0:
                  if (col == FIRSTCOL) break;
               default:  /* -2 or 1 */
                  ZXdowncurs();
            }
            return err;
      }
      if (onLF >= 0 || getexpr(&val_i,&val_s) != NOERR) return BADARG;

      if (doSTR)
         { onLF = !val_s[0]; err = ZXprintf(val_s); }
      else
         { ltoa(val_i, val_s); addpt(val_s);
           onLF = 0; err = ZXprintf(val_s); }

      free(val_s);
   }
}


/* adds decimal point to number */
addpt(str)
   char *str;
{
   int i, j;
   i = strlen(str);
           
   if (i == 1) { str[1] = str[0]; str[0] = '0'; str[2] = '\0'; i = 2;  }
   if (str[i-1] != '0')
      { /* shift_lo */ str[i+1] = '\0';  str[i] = str[i-1];
        shift_hi: str[i-1] = str[i-2]; str[i-2] = '.';
        if (i == 2)
           { for (j=4; j>0; j--) str[j] = str[j-1];
             str[0] = '0'; } }
   else
      { if (str[i-2] != '0') { str[i] = '\0'; goto shift_hi; }
        if (i == 2) str[1] = '\0'; else str[i-2] = '\0'; }
}


/* checks for no argument from command line */
noarg ()
{
   while (str[sp] == ' ') sp++;
   if (str[sp] != EOL)
      return BADCHAR;
   else
      return NOERR;
}


/* skips to end of line */
skipln ()
{
   while (str[sp] != EOL) sp++;
}


/* adds new loop nest for "var" */
nestlp (name,start,stop,step)
   char name[];
   long start, stop, step;
{
   char ch;
   addr mem, *var;
   long  *end, *inc;
   int  i, *ptr;

   if ((ch=name[strlen(name)-1])=='$' || ch=='(') return BADARG;

   i = 0;
   while (i < lp_cnt)
      if (!strcmp(lp_var[i],name)) break; else i++;

   /* if limit reached, resize loop structure */
   if (i == lp_len)
   {
      ptr=lp_ptr; end=lp_end; inc=lp_inc; var=lp_var;
      i=0; lp_len+=BLKNEST; initlp(CREATE);
      while (i<lp_cnt)
         { lp_ptr[i]=ptr[i]; lp_end[i]=end[i]; lp_inc[i]=inc[i]; lp_var[i]=var[i]; i++; }
      free(ptr); free(end); free(inc); free(var);
   }

   storevar(name,0L,start,"");

   if (step > 0L) if (start > stop) return NOVAL;
   if (step < 0L) if (start < stop) return NOVAL;

   lp_ptr[i] = sp;
   lp_end[i] = stop;
   lp_inc[i] = step;
   if (i == lp_cnt)
   {
      memarg(&mem,strlen(name)+1);
      lp_var[i] = mem; strcpy(mem,name);
      lp_cnt++; 
   }
      
   return NOERR;
}


/* skips to next increment of "var" */
nextlp (name)
   char name[];
{
   char ch;
   addr val_s;
   long val_i, step, stop;
   int  i;

   if ((ch=name[strlen(name)-1])=='$' || ch=='(') return BADARG;

   i = lp_cnt;
   while (i > 0)
      if (!strcmp(lp_var[--i],name))
      {
         fetchvar(name,0L,&val_i,&val_s);
         free(val_s); step = lp_inc[i];  stop = lp_end[i];
         val_i += step; storevar(name,0L,val_i,"");

         if (step > 0L) if (val_i > stop) return NOERR;
         if (step < 0L) if (val_i < stop) return NOERR;
         sp=lp_ptr[i]; return NOERR;
      }

   return BADARG;
}


/* list program */
listpgm (ed)
   int ed;
{
   int i, j, cp;
   char ch, lstr[LINELEN];

   dquote = 1; /* show double quote: "" */

   i=0; cp=sp;
   while (i<=(ed*10))   /* if editing, new line to middle of screen */
      { i++; if (cp > 1) { do cp--; while (pstr[cp-1]!=EOL); } else { cp=1; break; } }

   j = -1;
   while (pstr[cp]!=EOP)
   {
      if (j == -1)
      {
         j = 0;
         if (pstr[cp+1] == ' ')
            { lstr[j++] = ' '; goto in2; }  /* reduce code to indent 3 */
         else if (pstr[cp+2] == ' ')
            { in2: lstr[j++] = ' '; goto in3; } /* reduce code to indent 2  */
         else if (pstr[cp+3] == ' ')
            { in3: lstr[j++] = ' '; }
         i = j;
      }

      if (ed) if (cp == sp)
         { lstr[j++] = '%'; lstr[j++] = '>'; cp++; }

      lstr[j++] = pstr[cp];

      if ((ch=pstr[cp++]) == '\\')
         { lstr[j++] = pstr[cp++]; goto skipch; }
      if (ch == '%')
         { skipch: lstr[j++] = pstr[cp++]; }
  
      if (i != COLLEN) if (ch != EOL) { i++; continue; }

      if (i == COLLEN) if (ch == EOL) { j--; } lstr[j] = '\0';
      if (ZXprintf(lstr) != NOERR) return NOMORE;
      i = 1; j = 0 - (ch == EOL);

      trap3(key, ret);  /* get next key input */
      if (!ret[0]) if (ZXBREAK) break;
   }

   dquote = 0; /* show single quote: " */
   return NOERR;
}


/* saves program to file */
savefile (fname)
   char fname[];
{
   char ch;
   int fp, i;

   i=0;
   while ((ch=fname[i])!='\0') fname[i++] = tolower(ch); /* lowcase name */
   if ((fp=fopen(fname,"w")) == NULL) return BADARG;

   i=1;
   while ((ch=pstr[i++])!=EOP)
   {
      fputc(ch,fp);

      /* we change our intneral \"" to \" to be compatible */
      if (ch == '"') if (pstr[i] == '"') if (pstr[i-2] == '\\') i++;
   }
       
   fclose(fp);

   return NOERR;
}


/* loads file into string */
loadfile (fname)
   char fname[];
{
   int fp, l, ol, ln, ch, err;

   if ((fp=fopen(fname,"r")) == NULL) return BADARG;
 
   if (setpmem(fp) != NOERR) return NOMORE;

   l = 0; ln = 0; clearpgm();
   while ((err=checkline(fp,&ln,&l)) == NOERR)
      if ((err=readline(fp,&l)) != NOERR) break;

   if (err != BADCHAR)
      { if (pstr[l] != EOL) { setpstr(++l,EOL); } setpstr(++l,EOP); plen=l; }
   else
      { ol=l; readline(fp,&l); pgmerr("BAD LINE",ol); clearpgm(); }
   fclose(fp);
   return NOERR;
}


/* allocates file buffer for program */
setpmem (fp)
   int fp;
{
   long size;

   size = BLKLEN;
   while (fseek(fp, size, 0) == 0) size+=BLKLEN;
   fseek(fp, 0L, 0);

   if (mlen) free(pstr); mlen = size;
   memarg(&pstr, mlen);
   if (pstr == NULL) { mlen = 0; return NOMORE; }

   return NOERR;
}

  
/* reads character into program string, realloc's memory */
setpstr (l,ch)
   int l;
   char ch;
{
   char *ostr, *pmem;
   int  i;

   if (l >= mlen)
   {
      memarg(&pmem, l+BLKLEN);
      if (pmem == NULL) return;

      ostr = pstr; pstr = pmem;
      i=0; while (i<mlen) pstr[i] = ostr[i++];
      if (mlen) free(ostr);
      mlen = l + BLKLEN;
   }
   pstr[l] = ch;
}


/* checks line number from file */
checkline (fp,ln,l)
   int fp, *ln, *l;
{
   char lnum[NUMLEN];
   int  i, num, ch, err;

   i = 0;
   while (((ch=fgetc(fp)) == ' ') || (ch == EOL));
   if (ch == EOF) return EOF;
   while (isdigit(ch))
   {
      lnum[i++] = ch; setpstr(++(*l),ch);
      if ((ch=fgetc(fp)) == EOF) { (*l) -= i; return EOF; }
   }

   lnum[i] = '\0'; num = atoi(lnum); setpstr(++(*l),ch);
   if (ch == ' ') if (num > *ln) if (num <= MAXLNUM)
      { *ln = num; return NOERR; }
   
   return BADCHAR;
}


/* reads line from file */
readline (fp,l)
   int fp, *l;
{
   int ch;

   if (pstr[*l] != EOL)
   {
      while ((ch=toupper(fgetc(fp))) != EOL)
      {
         if (ch == '\\')     /* we change \" to our intneral \"" to be compatible */
            { setpstr(++(*l),ch); if ((ch=toupper(fgetc(fp))) == '"') setpstr(++(*l),ch); }
         if (ch == EOF) return EOF; 
         if (ch != '\r') setpstr(++(*l),ch);  /* get rid of Windows carriage returns */
      }
      setpstr(++(*l),EOL);
   }
   return NOERR;
}


/* jumps to line number */
seekline (ln)
   int ln;
{
   return findline(ln*100L);
}


/* gets line from program */
getline (lp, bp, ep)
   int lp, *bp, *ep;
{
   char ch;

   *bp=lp; while (*bp) if (pstr[(*bp)-1] != EOL) (*bp)--; else break;
   *ep=lp; while ((ch=pstr[*ep]) != EOP) if (ch != EOL) (*ep)++; else break;
}


/* gets line number from command line */
linenum (ln)
   int * ln;
{
   char ch, num[NUMLEN];
   int  i;

   i=0;
   while(str[sp] == ' ') sp++;
   while(isdigit(ch=str[sp])) { num[i++]=ch; sp++; }
   num[i] = '\0'; *ln = atoi(num);
   if (i) return NOERR; else return NOVAL;
}


/* inserts line into program */
insertline (ln)
   int ln;
{
   char ch;
   int err, cl, cp, np, bp, ep;

   if ((ln < 1) || (ln > MAXLNUM)) return BADARG;
   cp=sp; while (cstr[cp]==' ') cp++;

   /* first delete line if found */
   err=seekline(ln); getline(sp,&bp,&ep);
   if (err == NOERR)
      { np=bp; while (pstr[ep] != EOP) { setpstr(np++, pstr[++ep]); } plen=np-1; }

   ep=plen;

   /* if new line is not empty, add it */
   if (cstr[cp] != EOL)
   {
      cp=0; while (((ch=cstr[cp])==' ') || (ch=='0')) cp++;
      cl=strlen(cstr)-cp; while (ep >= bp) setpstr(ep+cl, pstr[ep--]);
      while ((ch=cstr[cp++])!='\0') setpstr(bp++, ch);
      plen+=cl;
   }

   return NOERR;
}


/* adds new line to program and lists it */
addline ()
{
   int ln, err;

   sp = 0; str = cstr;

   if (linenum(&ln) != NOERR)
      return NOVAL;

   if (strchr(" \n", str[sp]) != NULL) if (insertline(ln) == NOERR)
      { seekline(ln); listpgm(1); return NOERR; }

   return BADARG;
}


/* finds line number (rounds to nearest integer) or gives error */
findline (num)
   long num;
{
   int key, ln, beg, end, bp, ep;

   key = (num + 50L) / 100L;  /* fraction into int, but round */

   str=pstr; beg=1; end=plen;
   while (beg < end)
   {
      sp = (beg + end) >> 1;   /* divide by 2 */

      getline (sp, &bp, &ep); sp=bp;

      if (linenum(&ln) != NOERR)
         break;

      if (key == ln)
         return NOERR;

      if (key < ln)
         end = bp-1;
      else 
      {
         beg = ep+1;
         sp = beg;
      }
   }

   while (isdigit(str[sp])) sp++;

   return NOVAL;
}


/* edits line of program (rounds to nearest integer) */
editline (num)
   long num;
{
   int  i, err, bp, ep;

   findline(num);
   if (pstr[sp]==EOP) return NOERR;

   getline(sp,&bp,&ep);

   i = 0;
   while (bp < ep)
      { cstr[i++] = pstr[bp++];
        if (i > LINEMAX) return MAXLIM; }
   cstr[i] = '\0';

   ZXgets(cstr, 0);

   err = addline();
   if (err == NOVAL)
      { listpgm(1); err = NOERR; }
   reset();

   return err;
}


/* points to error in program line */
pgmerr (msg,lp)
   char msg[]; int  lp;
{
   char ch, estr[LINELEN];
   int  i, bp, ep;

   getline(lp,&bp,&ep);

   i = 0;
   while (bp < ep)
      { estr[i++] = pstr[bp++];
       if (lp == bp) { estr[i] = '\0'; lp = ZXlen(estr); }
       if (i > LINEMAX) break; }
   estr[i] = '\0';
   if (lp > i) lp = ZXlen(estr);

   printerr(msg,estr,lp);
}


/* points to error in command line */
showerr (msg)
   char msg[];
{
   dquote = 1;
   if (str == cstr)
      printerr(msg,"",sp);
   else
      pgmerr(msg,sp);
}


/* displays error message */
printerr (msg,estr,pt)
   char msg[], estr[]; int pt;
{
   int len;

   cls(stdout);
   row = FIRSTROW; col = FIRSTCOL;

   if (estr[0] != '\0')
   {
      ZXprintf(estr);
      ZXdowncurs();

      pt = pt % COLLEN;
      col = FIRSTCOL + 8 * pt;

      if ((pt-3) < (len=strlen(msg)))
         { ZXprintf("<--"); }
      else
         { col -= 24; ZXprintf("-->"); col -= 8 * (len + 3); }
   }
   ZXprintf(msg);
}


/* dynamic memory allocation of arg */
memarg (arg,len)
   addr *arg; int len;
{
   if (globerr==NOERR) if (len>0) if ((*arg=malloc(len)) != NULL) return;

   *arg = NULL;
   syserr(NOHEAP);
}


/* sets char ptr to ch */
setchp (chp,ind,ch)
   char chp[]; int ind; char ch;
{
   chp[ind] = ch;
}


/* gets ch from char ptr */
getchp (chp,ind)
   char chp[]; int ind;
{
   return chp[ind];
}


/* sets int ptr to num */
setintp (intp,ind,num)
   long intp[], ind, num;
{
   intp[ind] = num;
}


/* gets num from int ptr */
getintp (intp,ind,num)
   long intp[], ind, *num;
{
   *num = intp[ind];
}


/* gives system error */
syserr (err)
   int err;
{
   if (globerr == NOERR)
   {
      switch(globerr=err)
      {
         case NOHEAP : showerr("BAD HEAP QUERY"); break;
         default     : showerr("SYSTEM ERROR");
      }
      reset();
   }
}


/* reset input strings */
reset ()
{
   delim = spc; dquote = 0;
   cstr[0] = '\0'; str = cstr; sp = 0;
   rt_ind = -1; initlp(RESET);
   if (pstr[plen] != EOP ) syserr(UNKNOWN);
}


/* clear program */
clearpgm ()
{
   setpstr(0,EOL);
   setpstr(1,EOP);
   plen = 1;
}


/* set variables (either create new or clear old) */
initvar (clr)
   int clr;
{
   int i;

   if (!clr)
      { memarg(&var_nam, var_len*4);
        memarg(&var_ptr, var_len*4);
        memarg(&var_num, var_len*4); }

   i=0;
   while (i<var_len)
      { if (clr)
           { if (var_nam[i] != 0L) free(var_nam[i]);
             if (var_ptr[i] != 0L) free(var_ptr[i]); }
        var_nam[i] = 0L;
        var_ptr[i] = 0L;
        var_num[i] = 0L;
        i++; }
}


/* set loop structure (create new or clear old) */
initlp (clr)
   int clr;
{
   if (clr)
      while (lp_cnt > 0)
         { lp_cnt--;
           free(lp_var[lp_cnt]); lp_var[lp_cnt]=0L; }
   else
      { memarg(&lp_ptr, lp_len*2);
        memarg(&lp_end, lp_len*4);
        memarg(&lp_inc, lp_len*4);
        memarg(&lp_var, lp_len*4); }
}


/* hash function */
hash (var)
    char var[];
{
   int i, h;
   char ch;

   h=0; i=0;
   while ((ch=var[i++])!='\0') h+=ch;

   return (h % var_len);
}

