/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * cmd_c: parser for ZXSimulator, a ZX81 ROM emulator
 */


#define MAIN

#include stdio_h
#include cmd_h

main ()
{
   int err;

   initPARSE();

   while (1)
   {
      if (globerr != NOERR)
         { globerr = NOERR; clearpgm(); }

      reset();

      /* get next input from user */
      ZXgets(cstr, 0);

      /* list program on empty command */
      if (cstr[0] == EOL)
         { listpgm(0); continue; }

      /* add BASIC line: NOERR if ok, NOVAL if not, error otherwise */
      err = addline();
      if (err == NOERR) continue;

      /* otherwise we need to parse the command */
      if (err == NOVAL)
         err = PARSEcmds();
     
      /* and report any errors */
      switch (err)
      {
         case NOERR  :
         case NOMORE : break;
         case MAXLIM : showerr("MAX LIMIT"); break;
         case NOCMD  : showerr("NO COMMAND"); break;
         case BADARG : showerr("BAD ARGUMENT"); break;
         default     : showerr("UNKNOWN");
      }
   }
}


/* initializes the parser */
initPARSE ()
{
   initQLscr();
   seed = date();
   key[4] = get_chanid(stdin);

   var_len=VARSIZE;
   initvar(CREATE);

   rt_len=BLKNEST;
   memarg(&rt_ptr, BLKNEST*4);

   lp_len=BLKNEST;
   initlp(CREATE);

   clearpgm();
}


/* parser */
PARSEcmds()
{
   int  err, ln;

   for (;;)
   {
      if (ZXBREAK) return NOERR;

      if ((err=BASICcmds()) != NOERR) return err;

      if (str[sp] == EOL)        /* go to next line */
         { sp++; linenum(&ln); }
      if (str[sp] == EOP) return NOERR;
   }
}


/* BASIC commands */
BASICcmds ()
{
   char ch, *val_s, *vl2_s, *tmp_s, arg[CMDLEN], var[VARLEN];
   int err, i, j, k, l;
   long val_i, vl2_i, vl3_i, vl4_i, dim[MAXDIM];
   addr mem;

   if (getkey("CLEAR") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      initvar(RESET); initlp(RESET);
   }
   else if (getkey("CLS") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      cls(stdout);
      row = FIRSTROW; col = FIRSTCOL;
   }
   else if (getkey("DIM") == NOERR)
   {
      if (getvar(var) != NOERR) return BADARG;
      if ((ch=var[strlen(var)-1]) == '(') sp--;
      if (getsym("(") != NOERR) return BADARG;

      k = 1; l = 0;
      do { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           j = (val_i+50L)/100L;
           if (l < MAXDIM) dim[l++] = j; else return MAXLIM;
           if (j >= 1) k *= j; else return BADARG;
      } while (getsym(",") == NOERR);
      if (getsym(")") != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;

      if (findvar(var,&i) == NOVAL)
         { memarg(&mem,strlen(var)+1); var_nam[i]=mem; strcpy(mem,var); }
      else
         { free(var_ptr[i]); }

      if (ch == '$')
         { if (l > 2) return MAXLIM;
           memarg(&mem, k+1); if (l == 2) vl2_i = dim[1]; else vl2_i = 0L;
           j=0; while (j<k) { setchp(mem,j++,' '); } setchp(mem,j,'\0'); }
      else
         { memarg(&mem,(k+l+1)*4);
           vl2_i = l; setintp(mem,0L,vl2_i); vl2_i = 1L;
           while (vl2_i<=l) { setintp(mem,vl2_i,dim[vl2_i-1]); vl2_i++; }
           while (vl2_i<=k+l) setintp(mem,vl2_i++,0L); vl2_i = 0L; }

      var_ptr[i] = mem; var_num[i] = vl2_i;
   }
   else if (getkey("EDIT") == NOERR)
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      return editline(val_i);
   }
   else if (getkey("FOR") == NOERR)
   {
      if (getvar(var) != NOERR || getsym("=") != NOERR) return BADARG;
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("TO") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("STEP") != NOERR)
         vl3_i = 100L;
      else if (getexpr(&vl3_i,&val_s) == NOERR)
         free(val_s);
      else
         return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((err=nestlp(var,val_i,vl2_i,vl3_i)) == NOVAL)
         { while (str[sp] != EOF)
              if (str[sp++] == EOL)
                 { while ((ch=str[sp]) != ' ') { if (ch == EOF) break; else sp++; }
                   if (getkey("NEXT") == NOERR) if (getvar(arg) == NOERR) if (!strcmp(var, arg))
                      break; } }
      else
         { return err; }
   }
   else if (getkey("GOSUB") == NOERR)
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((++rt_ind) >= rt_len) 
         { memarg(&mem,(rt_len+BLKNEST)*4); i=0; while (i<rt_len) { vl2_i=i;setintp(mem,vl2_i,rt_ptr[i]); i++; }
           free(rt_ptr); rt_ptr=mem; rt_len += BLKNEST; }
      rt_ptr[rt_ind] = sp; findline(val_i);
   }
   else if (getkey("GOTO") == NOERR)
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      findline(val_i);
   }
   else if (getkey("IF") == NOERR)
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("THEN") != NOERR) return BADARG;
      if (noarg() == NOERR) return BADARG;
      if (!val_i) skipln();
   }
   else if (getkey("INPUT") == NOERR)
   {
      memarg(&vl2_s,LINELEN);
      i = INPUT; goto assign;
   }
   else if (getkey("LET") == NOERR)
   {
      i = ASSIGN; assign:
      if (getvar(var) != NOERR) return BADARG;
      switch (var[strlen(var)-1])
      {
         case '$':
            if (getsym("(") != NOERR) { ch='$'; goto get_val; } sp = op;
            if (fetchvar(var,0L,&val_i,&val_s) != NOERR) return BADARG;
            if (getrange(val_i,&vl2_i,&vl3_i,ZXlen(val_s)) != NOERR) goto free_on_err;
            if (i == ASSIGN)
               { if (getsym("=") != NOERR) goto free_on_err;
                 if (getexpr(&val_i,&vl2_s) != NOERR) goto free_on_err; }
            else
               { ZXgets(vl2_s,1); }

            i=0; j=1; l=strlen(val_s);
            while (j++<vl2_i) switch(getchp(val_s,i++)) { case '\\': i++; case '%': i++; }

            j=0; vl4_i=0L; vl3_i=vl3_i-vl2_i+1L; vl2_i=i; val_i=0L; k=0;
            while (vl4_i++<vl3_i)
               { switch(getchp(val_s,i++)) { case '\\': k--; i++; case '%': k--; i++; }
                 switch(getchp(vl2_s,j++))
                    { case '\0': j--; val_i++; break; case '\\': k++; j++; case '%': k++; j++; } }

            memarg(&tmp_s,l+1); strcpy(tmp_s,val_s); free(val_s);
            l+=k+1; memarg(&val_s,l);  vl3_i=j;
            j=0; while (j<vl2_i) setchp(val_s,j,getchp(tmp_s,j++));
            i=0; while (i<vl3_i) setchp(val_s,j++,getchp(vl2_s,i++));
            i=0; while (i++<val_i) setchp(val_s,j++,' ');
            while (l>=j) { setchp(val_s,l,getchp(tmp_s,l-k)); l--; }
            free(tmp_s); free(vl2_s);
            break;
         case '(':
            if (getindex (var,&vl4_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            if (vl4_i == 0L) { get_val: vl4_i = -1L; }
         default:
            if (i == ASSIGN)
               { if (getsym("=") != NOERR) return BADARG;
                 if (getexpr(&val_i,&val_s) != NOERR) return BADARG; }
            else
               { val_s = vl2_s; ZXgets(val_s,1);
                 if (ch != '$') {
                    i=0; while ((ch=val_s[i])!='.' && ch!='\0') i++;
                    for (j=0; j<2; j++)
                       if (isdigit(val_s[++i])) val_s[i-1]=val_s[i]; else val_s[i-1]='0';
                    val_s[i]='\0'; val_i=atol(val_s); } }
      }

      if (noarg() != NOERR) { free_on_err: free(val_s); return BADARG; }

      storevar(var,vl4_i,val_i,val_s); free(val_s);
   }
   else if (getkey("LIST") == NOERR)
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           i = sp; tmp_s = str; findline(val_i); }
      else
         { i = sp; tmp_s = str; sp = 1; }

      if (listpgm(0) != NOERR) return NOMORE;
      sp = i; str = tmp_s;
   }
   else if (getkey("LOAD") == NOERR)
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if ((err=noarg()) == NOERR) err = loadfile(val_s);
      free (val_s); if (err != NOERR) return BADARG;
      initvar(RESET); reset();
   }
   else if (getkey("NEW") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      clearpgm(); initvar(RESET); reset();
   }
   else if (getkey("NEXT") == NOERR)
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (nextlp(var) != NOERR) return BADARG;
   }
   else if (getkey("PAUSE") == NOERR)
   {
      if (getexpr(&val_i,&val_s) == NOERR) { val_i=(val_i+50L)/100L; free(val_s); } else return BADARG;
      if (noarg() != NOERR) return BADARG;
      i=0;
      while (i<val_i) { trap3(key,ret); if (!ret[0]) { if (i) break; getchar(); } pause(5); i+=5; }
   }
   else if (getkey("PLOT") == NOERR)
   {
      k = 0;
      plot_command:
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getsym(",") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      i = labs((val_i + 50L) / 100L); j = labs((vl2_i + 50L) / 100L);
      if (i > MAXPLOTX || j > MAXPLOTY) return BADARG;
      block(stdout, 4, 4, FIRSTCOL+(i<<2), LASTROW+4-(j<<2), k);
   }
   else if (getkey("UNPLOT") == NOERR)
   {
      k = 7;
      goto plot_command;
   }
   else if (getkey("PRINT") == NOERR)
   {
      return putformat();
   }
   else if (getkey("RAND") == NOERR)
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           if (val_i) { seed = ((val_i+50L)/100L); return NOERR; } }
      seed = date();
   }
   else if (getkey("REM") == NOERR)
   {
      skipln();
   }
   else if (getkey("RETURN") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      if (rt_ind < 0) return BADARG;
      sp = rt_ptr[rt_ind--];
   }
   else if (getkey("RUN") == NOERR)
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           reset(); findline(val_i); }
      else
         { reset(); str = pstr; }
      initvar(RESET);
   }
   else if (getkey("SAVE") == NOERR)
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if ((err=noarg()) == NOERR) err = savefile(val_s);
      free (val_s); if (err != NOERR) return BADARG;
   }
   else if (getkey("SCROLL") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      scroll(stdout, -8);
      block(stdout, MAXROW, 1, 2, 0, 7); /* clear the top line */
      row = LASTROW; col = FIRSTCOL;
   }
   else if (getkey("STOP") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      reset();
   }
   else if (getkey("QUIT") == NOERR)
   {
      if (noarg() != NOERR) return BADARG;
      abort(0);
   }
   else                                /* if no more command */
   {
      if (getkey("FAST") != NOERR)     /* list of commands we ignore */
      if (getkey("SLOW") != NOERR)
         return NOCMD;
   }

   return NOERR;
}

