/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * cmd_h: header file for ZXSimulator, a ZX81 ROM emulator
 */


#define ZXBREAK    keyrow(7)==2) if (keyrow(3)==8          /* ctl-s */
#define getexpr    evalbool    /* top expression in precedence tree */

#define addr       long

#define EOL       '\n'
#define EOP       '\0'
#define LINELEN    128  /* was 256 */
#define LINEMAX    125  /* 3 less from LINELEN for EOL/EOF or cursor */
#define CMDLEN     20
#define VARLEN     20
#define BLKLEN     64
#define NUMLEN     10
#define MAXDIM     6
#define MAXLNUM    9999
#define MAXINT     32768

#define BLKNEST    10
#define VARSIZE    16
#define CREATE     0
#define RESET      1

#define INPUT      0
#define ASSIGN     1

#define NOERR      0
#define NOCMD     -1
#define NOVAL     -2
#define NOHEAP    -3
#define MAXLIM    -4
#define BADARG    -5
#define BADCHAR   -6
#define NOMORE    -7
#define UNKNOWN   -8

/* constant for PI */
#define PI         314L  /* (3.1415927) */

/* constants for random number generator */
#define P1         31421L
#define P2         6927L
#define N1         65535L   /* we bitwise-and 2^16 -1 to get new seed */
#define N2         65536L   /* range should be 0 - .99 so N1/N2=.9999 */

/* constants for ZX graphics */
#define WINTOP     1
#define WINSIDE    124    /* (512-(MAXCOL+4))/2 */
#define WINWIDTH   264    /* MAXCOL+4 */
#define WINHEIGHT  192    /* MAXROW+4 */
#define FIRSTROW   2
#define MAXROW     188
#define LASTROW    170
#define CURSROW    180
#define FIRSTCOL   2
#define MIDCOL     130
#define LASTCOL    250
#define MAXCOL     260
#define COLLEN     32
#define MAXPLOTY   43
#define MAXPLOTX   63

#ifdef MAIN

/* character map */
char sym[17] = {'"','#','$',':','?','(',')','>','<','=','+','-','*','/',';',',','.'};

/* global command variables */
int  globerr = NOERR;
char lst[14] = {' ',',',';','\n','(',')','>','<','=','+','-','*','/','\0'};
char spc[5] = {' ',',',';','\n','\0'};
char *delim;
int  doSTR;
char cstr[LINELEN] = "";
char *pstr;
int  mlen = 0;
int  plen;
char *str;
int  sp;
int  op;

/* data sturctures for variables */
addr *var_nam;
addr *var_ptr;
long *var_num;
int  var_len;

/* data structures for loops gosubs, and if statements */
int  *lp_ptr;
long *lp_end;
long *lp_inc;
addr *lp_var;
int  lp_len;
int  lp_cnt = 0;

long *rt_ptr;
int  rt_len;
int  rt_ind = -1;

long key[8] = {0X0};
long ret[8];

/* random seed */
long seed;

/* data structurs for ZX graphics */
int  col = FIRSTCOL;
int  row = FIRSTROW;
int  endrow = LASTROW;
int  dquote;
int  inprow;

#else

extern char sym[];
extern int  globerr;
extern char lst[];
extern char spc[];
extern char *delim;
extern int  doSTR;
extern char cstr[];
extern char *pstr;
extern int  mlen;
extern int  plen;
extern char *str;
extern int  sp;
extern int  op;
extern addr *var_nam;
extern addr *var_ptr;
extern long *var_num;
extern int  var_len;
extern int  *lp_ptr;
extern long *lp_end;
extern long *lp_inc;
extern addr *lp_var;
extern int  lp_len;
extern int  lp_cnt;
extern long *rt_ptr;
extern int  rt_len;
extern int  rt_ind;
extern long key[8];
extern long ret[8];
extern long seed;
extern int  col;
extern int  row;
extern int  endrow;
extern int  dquote;
extern int  inprow;

#endif

