/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * expr_c: expression builder for ZXSimulator, a ZX81 ROM emulator
 */


#include stdio_h
#include cmd_h

/* gets a variable from command line */
getvar (var)
   char var[];
{
   int  i;
   char ch;

   while (str[sp] == ' ') sp++;
   if (!isalpha(str[sp])) return NOVAL;
   i=0;
   while (i<VARLEN)
      if (isalpha(ch=str[sp]) || isdigit(ch) || (ch==' '))
         { if (ch != ' ')
              { var[i++] = ch; sp++; }
           else 
              { while (str[sp]==' ') sp++;
                switch (str[sp])
                   { case 'T':
                       if (getkey("THEN") == NOERR) { sp = sp - 4; goto end_var; }
                       if (getkey("TO") == NOERR) { sp = sp - 2; goto end_var; }
                       break;
                    case 'S':
                       if (getkey("STEP") == NOERR) { sp = sp - 4; goto end_var; }
                       break;
                    case 'A':
                       if (getkey("AND") == NOERR) { sp = sp - 3; goto end_var; }
                       break;
                    case 'O':
                       if (getkey("OR") == NOERR) { sp = sp - 2; goto end_var; }
         }    }    }
      else
         { if (ch == '$' || ch == '(')
              { var[i++] = ch; sp++; if (i == VARLEN) break; }
           end_var: var[i] = '\0'; return NOERR; }

   return BADARG;
}


/* finds var in hashtable */
findvar(var,vp)
   char var[]; int *vp;
{
   int i, j;
   long *num;
   addr *nam, *ptr;

   i = hash(var);
   
   j=i;
   do {
      if (var_nam[j] == 0)
         { *vp = j; return NOVAL; }
      else if (!strcmp(var_nam[j],var))
         { *vp = j; return NOERR; }
      j=(j+1)%var_len;
   } while (j!=i);


   /* if hash is filled up we double space */
   nam=var_nam; ptr=var_ptr; num=var_num;

   var_len+=VARSIZE;
   initvar(CREATE);

   i=0;
   while (i<(var_len-VARSIZE))
      { findvar(nam[i],&j);
        var_nam[j]=nam[i]; var_ptr[j]=ptr[i]; var_num[j]=num[i]; i++; }
   free(nam); free(ptr); free(num);

   return findvar(var,vp);
}


/* sets a variable */
storevar (var,ind,val_i,val_s)
   char var[]; long ind, val_i; char val_s[];
{
   addr mem;
   char ch;
   int  i, err;

   if ((err=findvar(var,&i)) == NOVAL)
      { memarg(&mem,strlen(var)+1);
        var_nam[i]=mem; strcpy(mem,var); }

   if ((ch=var[strlen(var)-1]) == '$')
      { memarg(&mem,strlen(val_s)+1);
        if (err == NOERR) free(var_ptr[i]);
        var_ptr[i] = mem; strcpy(mem,val_s); 
        if (ind < 0) var_num[i] = 0L; }
   else if (ch == '(')
      setintp(var_ptr[i],ind,val_i);
   else
      var_num[i] = val_i;

   return NOERR;
}


/* retrievs value of a variable */
fetchvar (var,ind,val_i,val_s)
   char var[]; long ind, *val_i; addr *val_s;
{
   char ch;
   char *ptr;
   int i;

   if (findvar(var,&i) != NOERR) return BADARG;

   if ((ch=var[strlen(var)-1]) == '$')
      { memarg(val_s,strlen(ptr=var_ptr[i])+1); strcpy(*val_s,ptr);
        *val_i = var_num[i]; }
   else 
      { if (ch == '(')
           getintp(var_ptr[i],ind,val_i);
        else
           *val_i = var_num[i];
        memarg(val_s,NUMLEN); setchp(*val_s,0,'\0'); }

   return NOERR;
}


/* retrievs array index for DIM'ed array */
getindex (var,val_i,val_s)
   char var[]; long *val_i; addr *val_s;
{
   int i;
   long j, k, ind, len, dim;

   if (findvar(var,&i) != NOERR) return BADARG;

   j = 1L; ind = 0L; getintp(var_ptr[i],0L,&dim);
   while (1) {
      if (getexpr(val_i,val_s) == NOERR) { k=(*val_i+50L)/100L; free(*val_s); } else return BADARG;
      getintp(var_ptr[i],j++,&len);
      if (k < 1 || k > len) return BADARG;
      ind = ind * len + k - 1L;
      if (j > dim) break;
      if (getsym(",") != NOERR) return BADARG;
   }
   if (getsym(")") != NOERR) return BADARG;

   *val_i = (1L + dim) + ind;
   memarg(val_s,NUMLEN); setchp(*val_s,0,'\0');

   return NOERR;
}


/* gets number from command line */
getnum (val_i,val_s)
   long *val_i; addr *val_s;
{
   char ch;
   int i;

   while (str[sp] == ' ') sp++;
   if (!isdigit(str[sp])) return NOVAL;
   i=0; memarg(val_s,NUMLEN);
   while (isdigit(ch=str[sp]))
      { setchp(*val_s,i++,ch); sp++; }
   if (str[sp]=='.')
      { if (isdigit(ch=str[++sp])) { setchp(*val_s,i++,ch); sp++; } else return BADARG;
        if (isdigit(ch=str[sp]))
           { setchp(*val_s,i++,ch); while (isdigit(str[++sp])); } else setchp(*val_s,i++,'0'); }
   else
      { setchp(*val_s,i++,'0'); setchp(*val_s,i++,'0'); }
   setchp(*val_s,i,'\0');
   *val_i = atol(*val_s); setchp(*val_s,0,'\0');

   return NOERR;
}


/* gets quoted string from command line */
getstr (val_s)
   addr *val_s;
{
   char ch;
   int i, tp, len;

   while (str[sp] == ' ') sp++;
   if (str[sp] == '"') tp = ++sp; else return NOVAL;
   while((ch=str[tp])!='"')
      if (ch=='\\') tp+=3; else if (ch=='%') tp+=2; else { tp++; if (ch==EOL) return BADCHAR; }
   len = tp-sp; memarg(val_s,len+1);
   i=0; while (i<len) setchp(*val_s,i++,str[sp++]);
   setchp(*val_s,i,'\0'); sp++;

   return NOERR;   
}


/* parses substring expression */
getsubstr (tmp_i,tmp_s,val_i,val_s)
   long tmp_i; addr tmp_s;
   long *val_i; addr *val_s;
{
   char ch;
   long from, to;
   int i, j, cp, in, idx, len;

   switch (getrange(tmp_i,&from,&to,(len=ZXlen(tmp_s))))
      { case NOVAL:
           *val_i=0L; *val_s=tmp_s; return NOERR;
        case BADARG:
           free(tmp_s); return BADARG; }

   if (to < from)
      { j=0; memarg(val_s,1); }
   else
      {
        cp=0;
        while (cp<2)  /* first pass see how much space, no copy */
        {
           if (cp) memarg(val_s,j+1);

           i=0; j=0; in=0; idx=0;
           while (idx<len)
           {
              idx++;
              switch (ch=getchp(tmp_s,i))
              {
                 case '%':
                    if (in || from == idx) 
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++; }
                         else
                            { i=i+2; j=j+2; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+2;
                    break;
                 case '\\':
                    if (in || from == idx)
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++;
                              setchp(*val_s,j++,getchp(tmp_s,i)); i++; }
                         else
                            { i=i+3; j=j+3; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+3;
                    break;
                 default:
                    if (in || from == idx)
                       { if (cp) setchp(*val_s,j,ch);
                         j++;
                         if (idx == to) goto end_idx;
                         in = 1; }
                    i++;
              }
           }
           end_idx:
           cp++;
        }
      }

   setchp(*val_s,j,'\0'); *val_i=0L;
   free(tmp_s);

   return NOERR;
}


/* gets range of string */
getrange (size,from,to,len)
   long size, *from, *to; int len;
{
   char *tmp_s;
   long row;
   int all, noTO;

   if (getsym("(") != NOERR) { if (size) { return BADARG; } return NOVAL; }

   all = 0; row = 0L;
   if (size)
      { if (getexpr(&row,&tmp_s) == NOERR) { row=(row+50L)/100L; free(tmp_s); } else return BADARG;
        if (getsym(",") != NOERR)
           { if (getsym(")") == NOERR) { all=1; sp--; } else return BADARG; }
        len = row * size; row = len - size; }

   if (noTO=(getkey("TO") != NOERR))
      { if (getexpr(from,&tmp_s) == NOERR) { *from=(*from+50L)/100L; free(tmp_s); } else goto set_from; }
   else
      { set_from: *from = 1L; }

   *from += row;

   if (!noTO || (getkey("TO") == NOERR))
      { if (getexpr(to,&tmp_s) == NOERR) { *to=(*to+50L)/100L+row; free(tmp_s); } else *to = len; }
   else
      { if (all) *to = *from + size - 1L; else *to = *from; }

   if (getsym(")") != NOERR) return BADARG;

   if ((*from < 1) || (*to < 0)) return BADARG;
   if (*to < *from) { *from=0L; *to=-1L; } else if ((*from > len) || (*to > len)) return BADARG;

   return NOERR;
}


/* gets value from command line */
getval (val_i,val_s)
   long *val_i; addr *val_s;
{
   addr tmp_s;
   long tmp_i;
   int  i, err;
   char var[VARLEN];

   if (getsym("(") == NOERR)  /* eval subexpression */
      { if (getexpr(val_i,val_s) != NOERR) return BADARG;
        return getsym(")"); }
   else if ((err=evalfun(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getnum(val_i,val_s)) != NOVAL)
      { doSTR = 0; return err; }
   else if ((err=getstr(&tmp_s)) != NOVAL)
      { if (err != NOERR) return err;
        err = getsubstr(0L,tmp_s,val_i,val_s);
        doSTR = 1; return err; }
   else if (getvar(var) == NOERR)
      { switch (var[strlen(var)-1])
           { case '$':
                if (fetchvar(var,0L,&tmp_i,&tmp_s) != NOERR) break;
                err = getsubstr(tmp_i,tmp_s,val_i,val_s);
                doSTR = 1; return err;
             case '(':
                if (getindex(var,&tmp_i,&tmp_s) != NOERR) break;
                free(tmp_s);
             default:
                doSTR = 0;
                return fetchvar(var,tmp_i,val_i,val_s); } }

   return BADARG;
}


/* gets {chr$(),len()} functions and unary {-,not} expressions */
evalfun (val_i,val_s)
   long *val_i; addr *val_s;
{
   int i, j;

   delim = lst;  /* function delimeter different from other keywords */
   if (getsym("-") == NOERR) {
      if (getval(val_i,val_s) != NOERR) return BADARG;
      *val_i = -*val_i;
      goto number;
   }
   switch (str[sp])
   {
      case 'A':
         if (getkey("ABS") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = labs(*val_i);
            goto number;
         }
         goto noval;
      case 'C':
         if (getkey("CHR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) { return BADARG; } free(*val_s);
            i = (*val_i + 50L) / 100L;
            if (i >> 8) return BADARG;

            j = i > 127; memarg(val_s,2+j);
            if (j) { i = i - 128; setchp(*val_s,0,'%'); }

            if (i == 0)
               i = 32;
            else if (i < 11 || i > 63)
               { i = 15; j=0; }
            else if (i >= 38 )
               i = i + 27;  /* 'A' is 65 for ASCII - 38 for ZX81 */
            else if (i >= 28)
               i = i + 20;  /* '0' is 48 for ASCII - 28 for ZX81 */
            else
               i = sym[i-11];

            setchp(*val_s,j++,i); setchp(*val_s,j,'\0');
            goto string;
         }
         if (getkey("CODE") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            i = getchp(*val_s,0);
      
            if (i == '\0')
               *val_i = 0L;
            else if (i == '\\')
               { if (getchp(*val_s,1)==':' && getchp(*val_s,2)==':') i=128; else i=67; }
            else
               { if (i != '%') j = 0; else { j = 128; i = getchp(*val_s,1); }

                 if (i == 32)
                    i = j;
                 else if (i >= 48 && i <= 57)
                    i = i - 20 + j;   /* '0' is 48 for ASCII - 28 for ZX81 */
                 else if (i >= 65 && i <= 90)
                    i = i - 27 + j;   /* 'A' is 65 for ASCII - 38 for ZX81 */
                 else if (i < 192)    /* 192 & up are QL control characters we want */
                    { *val_i = j; j = 0;
                      while (j < 17) if (sym[j] == i) break; else j++;
                      if (j == 17) i = 67; else i = j + 11 + *val_i; } }
            *val_i = i * 100L;
            goto number;
         }
         goto noval;
      case 'I':
         if (getkey("INKEY$") == NOERR) {
            i = 0; memarg(val_s,2); trap3(key, ret);
            if (!ret[0]) { setchp(*val_s,i++,toupper(getchar())); } setchp(*val_s,i,'\0');
            if (getchp(*val_s,0)==32) { setchp(*val_s,0,252); }  /* for space 32 map to 252 */
            goto string;
         }
         if (getkey("INT") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = *val_i / 100L * 100L;
            goto number;
         }
         goto noval;
      case 'L':
         if (getkey("LEN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = ZXlen(*val_s) * 100L;
            goto number;
         }
         goto noval;
      case 'N':
         if (getkey("NOT") == NOERR) {
            if (getexpr(val_i,val_s) != NOERR) return BADARG;
            *val_i = !*val_i;
            goto number;
         }
         goto noval;
      case 'P':
         if (getkey("PI") == NOERR) {
            *val_i = PI;
            goto new;
         }
         goto noval;
      case 'R':
         if (getkey("RND") == NOERR) {
            seed = (seed * P1 + P2) & N1;
            *val_i = (seed * 100L) / N2;
            goto new;
         }
         goto noval;
      case 'S':
         if (getkey("SGN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = lsign(*val_i) * 100L;
            goto number;
         }
         if (getkey("STR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            ltoa(*val_i,*val_s); addpt(*val_s);
            goto string;
         }
         goto noval;
      case 'V':
         if (getkey("VAL") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = atol(*val_s) * 100L;
            goto number;
         }
      default:
         noval: delim = spc; return NOVAL;
   }

   new:    memarg(val_s,NUMLEN);
   number: doSTR = 0; setchp(*val_s,0,'\0'); delim = spc; return NOERR;
   string: doSTR = 1; *val_i = 0L; delim = spc; return NOERR;
}


/* parses ** expressions */
evalpow (left_i,left_s)
   long *left_i; addr *left_s;
{
   int i;
   long right_i, pow;
   addr right_s;

   if (getval(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' :
            if (str[sp+1] != '*') return NOERR;
            sp = sp + 2;
            if (getval(&right_i,&right_s) != NOERR) goto err_pow;
            if (right_i % 10) goto err_pow;  /* power has to be whole numbers for now */
            i = 1; pow = *left_i;
            while (i++ < (right_i / 100L))
               pow = (pow * *left_i) / 100L;

            *left_i = pow;
            break;
         default :
            return NOERR;
      }

      doSTR = 0; free(right_s);
   }

   err_pow: free(*left_s); return BADARG; 
}


/* parses {*,/} expressions */
evalmul (left_i,left_s)
   long *left_i; addr *left_s;
{
   long right_i;
   addr right_s;

   if (evalpow(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' :
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) goto err_mul;
            *left_i = (*left_i * right_i) / 100L;
            break;
         case '/' :
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) goto err_mul;
            *left_i = (*left_i * 100L) / right_i;
            break;
         default  :
            return NOERR;
      }

      doSTR = 0; free(right_s);
   }

   err_mul: free(*left_s); return BADARG;
}


/* parses {+,-} expressions */
evaladd (left_i,left_s)
   long *left_i; addr *left_s;
{
   int i, j;
   long right_i;
   addr right_s, tmp_s;

   if (evalmul(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      j=0; 
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '+' :
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) goto err_add;

            /* concatenate string if we grabbed one */
            if (doSTR)
               { memarg(&tmp_s,strlen(*left_s)+strlen(right_s)+1);
                 i=0; while (getchp(*left_s,i)!='\0') setchp(tmp_s,j++,getchp(*left_s,i++));
                 i=0; while (getchp(right_s,i)!='\0') setchp(tmp_s,j++,getchp(right_s,i++));
                 setchp(tmp_s,j,'\0'); free(*left_s); *left_s = tmp_s;
                 *left_i = atol(*left_s); }
            else
               *left_i = *left_i + right_i;
            break;
         case '-' :
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) goto err_add;
            *left_i = *left_i - right_i;
            doSTR = 0;
            break;
         default  :
            return NOERR;
      }

      free(right_s);
   }

   err_add: free(*left_s); return BADARG;
}


/* parses {=,>,<,<>,<=,>=} expressions */
evaleql (left_i,left_s)
   long *left_i; addr *left_s;
{
   long right_i;
   addr right_s;

   if (evaladd(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getsym("=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) == 0);
         else
            *left_i = *left_i == right_i;
      }
      else if (getsym("<=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) <= 0);
         else
            *left_i = *left_i <= right_i;
      }
      else if (getsym("<>") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) != 0);
         else
            *left_i = *left_i != right_i;
      }
      else if (getsym("<") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) < 0);
         else
            *left_i = *left_i < right_i;
      }
      else if (getsym(">=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) >= 0);
         else
            *left_i = *left_i >= right_i;
      }
      else if (getsym(">") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) goto err_eql;
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) > 0);
         else
            *left_i = *left_i > right_i;
      }
      else
         return NOERR;

      if (doSTR) setchp(*left_s,0,'\0');
      doSTR = 0; *left_i *= 100L;  free(right_s);
   }

   err_eql: free(*left_s); return BADARG;
}


/* parses {and,or} expressions */
evalbool (left_i,left_s)
   long *left_i; addr *left_s;
{
   int leftSTR;
   long right_i;
   addr right_s;

   if (evaleql(left_i,left_s) != NOERR) return BADARG;
   leftSTR = doSTR;
   while (1)
   {
      if (getkey("AND") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) goto err_left;
         if (doSTR) goto err_both;            /* ZX81 BASIC: can't have string in right arg */
         if (!right_i) *left_i = 0L;                       /* returns left arg if both true */
         if (leftSTR) setchp(*left_s,0,'\0');                      /* false is empty string */
      }
      else if (getkey("OR") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) goto err_left;
         if (leftSTR || doSTR) goto err_both;   /* ZX81 BASIC: can't have string in OR args */
         if (right_i) *left_i = 100L;             /* returns 1 if right true, else left arg */
      }
      else
         return NOERR;

      leftSTR=0; free(right_s);
   }

   err_both: free(right_s);
   err_left: free(*left_s);
   return BADARG;
}

