/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * graph_c: graphics routines for ZXSimulator, a ZX81 ROM emulator
 */


#include stdio_h
#include cmd_h

/* initializes the display */
initQLscr ()
{
   window(stdout,WINWIDTH,WINHEIGHT,WINSIDE,WINTOP);
   paper(stdout,7);
   ink(stdout,0);
   border(stdout,1,0);
   mode(4);
}


/* gets input ZX81 style */
ZXgets (lstr, io)
   char lstr[]; int io;
{
   int i, j, k, r, c, in;
   char ch, curs, cl[LINELEN];

   dquote = 1;        /* show double quote: "" */
   endrow = CURSROW;  /* set end row to edit mode */
   inprow = CURSROW;

   k=0; in=0;
   if (io)
   {
      r = row; c = col; curs = 'L';
      cl[k++] = '%'; cl[k++] = curs;
   }
   else
   {
      curs = 'K';
      if (lstr[0] == '\0')
      {
         cl[k++] = '%'; cl[k++] = curs;
      }
      else
      {
         j=0;
         while ((ch=lstr[j])!='\0')
         {
            if (!in) if (ch == ' ') { in = k; cl[k++] = '%'; cl[k++] = curs; }
            
            cl[k++] = ch; j++;
         }
      }
   }
   cl[k++] = '\0';
         
   while (1)
   {
      ZXclear(cl, 0);
      ZXprintf(cl);
      ZXprintf("  ");  /* two trailing spaces to remove prev char w/deleting */

      ch = toupper(getchar());

      if (ch == 200)
      {
         if (in < (k-5) && cl[in+2] == '\\')
            { cl[in] = '\\'; cl[in+1] = cl[in+3]; cl[in+2] = cl[in+4];
              cl[in+3] = '%'; cl[in+4] = curs; in = in + 3; }
         else if (in < (k-4) && cl[in+2] == '%')
            { cl[in+1] = cl[in+3]; cl[in+3] = curs; in = in + 2; }
         else if (in < (k-3))
            { cl[in] = cl[in+2]; cl[in+1] = '%'; cl[in+2] = curs; in = in + 1; }
      }
      else if (ch == 194)
      {
         if (in > 2 && cl[in-3] == '\\')
            { i=in-3; while (i<(k-2)) { cl[i] = cl[i+3]; i++; } in = in - 3; k = k - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { i=in-2; while (i<(k-1)) { cl[i] = cl[i+2]; i++; } in = in - 2; k = k - 2;}
         else if (in > 0)
            { i=in-1; while (i<k) { cl[i] = cl[i+1]; i++; } in = in - 1; k = k - 1; }
      }
      else if (ch == 192)
      {
         if (in > 2 && cl[in-3] == '\\')
            { cl[in+1] = cl[in-1]; cl[in] = cl[in-2]; cl[in-1] = '\\';
              cl[in-3] = '%'; cl[in-2] = curs; in = in - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { cl[in+1] = cl[in-1]; cl[in-1] = curs; in = in - 2; }
         else if (in > 0)
            { cl[in+1] = cl[in-1]; cl[in-1] = '%'; cl[in] = curs; in = in - 1; }
      }
      else if (ch == 19)   /* Ctrl-S quits editing (returns empty) */
      {
         if (io || k != 3) 
            { k=0; goto end_input; }
         continue;
      }
      else if (ch == 10)
      {
         i=0; j=0; k=0;
         while ((ch=cl[j]) != '\0')
         {
            if (j++ == in) { j++; continue; }

            if (!k) if( ch == ' ') continue;        /* skip all leading spaces */

            lstr[i++] = ch;
            if (ch != ' ') k = i;  /* keep track of first trailing space or end */
         }

         end_input:
         lstr[k] = EOL; /* add end with new-line */
         lstr[k+1] = '\0';

         break;
      }
      else 
      {
         if (k > LINEMAX) continue;

         if (ch != '\\') if (ch != '%')
            { i=k; while (i>=(in+1)) cl[i--] = cl[i];
              cl[in++] = ch; k = k + 1;
              continue; }

         cl[in+1] = 'G';
         ZXclear(cl, 0);
         ZXprintf(cl);
         
         if (ch == '\\')
            { i=k+2; while (i>=(in+3)) { cl[i] = cl[i-3]; i--; } cl[in++] = '\\';
              cl[in++] = toupper(getchar()); cl[in++] = toupper(getchar());
              k = k + 3; }
         else
            { i=k+1; while (i>=(in+2)) { cl[i] = cl[i-2]; i--; } cl[in++] = '%'; 
              cl[in++] = toupper(getchar()); k = k + 2; }

         cl[in+1] = curs;
      }
   }

   if (io)
   {
      lstr[k] = '\0';
      ZXclear(cl, 1);
      row = r; col = c;
   }
   else
   {
      /* if we are 'E'diting then no cls */
      i=0; while (lstr[i]==' ') i++;
      if (lstr[i] != 'E') cls(stdout); else ZXclear(cl, 1);

      row = FIRSTROW; col = FIRSTCOL;
   }

   dquote = 0;        /* back to single quote: "      */
   endrow = LASTROW;  /* reset end row to screen mode */
}


/* gets number of ZX characters */
ZXlen(lstr)
   char lstr[];
{
   int i, l;
   char ch;

   i = 0; l = 0;
   while ((ch=lstr[i++])!='\0')
      { switch(ch) { case '\\': if (lstr[i]=='"') { l+=dquote; } i++; case '%': i++; } l++; }

   return l;
}


/* clears edit window */
ZXclear (lstr, all)
   char lstr[]; int all;
{
   int i, j, r;


   r = CURSROW - (((j=ZXlen(lstr)) / COLLEN) * 8);
   if ((r < inprow) || all)
   {
      block(stdout, MAXCOL, 2, 2, r-2, 7); /* clear the top 2 lines */

      inprow = r;
      if (!all) j = (1 + ((CURSROW - r) / 8)) * COLLEN;

      i=0; row = r; col = FIRSTCOL - 8;
      while (i++<j)
      {
         ZXadjustcurs();
         ZXputc(' ', 0, 7);
      }
   }
   else
      r = inprow;

   row = r; col = FIRSTCOL;
}


/* prints in ZX81 font */
ZXprintf (lstr)
   char lstr[];
{
   char ch;
   int i;

   i=0;
   while (1)
   {
      ch=lstr[i++];
      if (row > endrow) if (i==1 || ch!='\0') return NOMORE;
      switch (ch)
      {
         case '\0':
            return NOERR;
         case EOL:
            ZXdowncurs();
            continue;
         case '\\':
            ZXgraphics(lstr[i++], lstr[i++]);
            break;
         case '%':
            ZXputc(lstr[i++], 7, 0);
            break;
         default:
            ZXputc(ch, 0, 7);
      }
      ZXadjustcurs();
   }
}


/* ZX81 font character */
ZXputc (ch, ink, pap)
   char ch; int ink, pap;
{
   int i, p1, p2;
   char* pt;

   block(stdout, 8, 8, col, row, pap);

   switch (ch)
   {
      /* LETTERS */

      case 'A':
         pt = "aQMJMraT";
         break;
      case 'B':
         pt = "iIMJIraSJtaV";
         break;
      case 'C':
         pt = "aQLJIrIuaV";
         break;
      case 'D':
         pt = "aILJIjJsImaN";
         break;
      case 'E':
         pt = "iVqIMJaS";
         break;
      case 'F':
         pt = "qIMJaS";
         break;
      case 'G':
         pt = "YdaQLJIrIuaV";
         break;
      case 'H':
         pt = "NIaSNq";
         break;
      case 'I':
         pt = "iQLbiV";
         break;
      case 'J': /* adds extra block call 3 vs 4 */
         pt = "JLaVMq";
         break;
      case 'K':
         pt = "NIIiIbQSIdImIv";
         break;
      case 'L':
         pt = "NIiV";
         break;
      case 'M':
         pt = "IRIjQ[NINq";
         break;
      case 'N':
         pt = "IRI[IdImNINq";
         break;
      case 'O':
         pt = "aQLJLraV";
         break;
      case 'P':
         pt = "iIMJaTJr";
         break;
      case 'Q':
         pt = "aQLJLraVI\\Ie";
         break;
      case 'R':
         pt = "iIMJaTJrImIv";
         break;
      case 'S':
         pt = "aQIJIMaSJtaV";
         break;
      case 'T':
         pt = "yAMZ";
         break;
      case 'U':
         pt = "KIJLaVMq";
         break;
      case 'V':
         pt = "LILqIUImQ^";
         break;
      case 'W':
         pt = "MIMqQ]IVIn";
         break;
      case 'X':
         pt = "IIIqIRIjR[IUImINIv";
         break;
      case 'Y':
         pt = "IAIqIJIjISIcK\\";
         break;
      case 'Z':
         pt = "qIIjIcI\\IUqN";
         break;

      /* NUMBERS */

      case '0':
         pt = "aQLJLrIjaVIcI\\IU";
         break;
      case '1':
         pt = "QYIRLbiV";
         break;
      case '2':
         pt = "aQIJJraTIMqN";
         break;
      case '3':
         pt = "aQIJIrQcJtIMaV";
         break;
      case '4':
         pt = "NaIZISILqM";
         break;
      case '5':
         pt = "qIIJIMaSJtaV";
         break;
      case '6':
         pt = "aQLJaSJtaV";
         break;
      case '7':
         pt = "qIIrIkIdJ]";
         break;
      case '8':
         pt = "iIIJIraSJLJtaV";
         break;
      case '9':
         pt = "aQJJLraTaV";
         break;

      /* SYMBOLS */

      case '"':
         pt = "JYJq";
         break;
      case '#':
         pt = "YYLRIraKqN";
         break;
      case '$':
         pt = "OaiRISiTIuiV";
         break;
      case ':':
         pt = "I[I^";
         break;
      case '(':
         pt = "IiLbIn";
         break;
      case ')':
         pt = "IQLZIV";
         break;
      case '>':
         pt = "IRI[IdI]IV";
         break;
      case '<':
         pt = "IjIcI\\IeIn";
         break;
      case '=':
         pt = "iSiU";
         break;
      case '*':
         pt = "KcIZiTIjI^In";
         break;
      case '-':
         pt = "iT";
         break;
      case '+':
         pt = "MbiT";
         break;
      case '_':
         pt = "iV";
         break;
      case '/':
         pt = "IrIkIdI]IV";
         break;
      case ';':
         pt = "IZJ]IW";
         break;
      case ',':
         pt = "JeI_";
         break;
      case '.':
         pt = "R]";
         break;
      case ' ':
         return;

      /* UNKOWN is ? */

      default:
         pt = "aQIJIrIkIdIf";
   }

   while (pt[0])
   {
      p1 = pt[0];
      p2 = pt[1];

      block(stdout, p1 >> 3 & 7, p1 & 7, col+(p2 >> 3 & 7), row+(p2 & 7), ink);

      pt = pt + 2;
   }
}


/* prints ZX81 graphics character */
ZXgraphics (ch1, ch2)
   char ch1, ch2;
{
   switch (ch1)
   {
      case '"':
         if (ch2 == '"')
            { ZXputc ('"', 0, 7);
              if (dquote)
                 { ZXadjustcurs();
                   if (row <= endrow)
                      ZXputc ('"', 0, 7); }
              return; }
         break;
      case '#':
         if (ch2 == '#')  /* 248 odd stipple */
            { ZXgrey(248, 248); return; }
         break;
      case '~':
         if (ch2 == '~')  /* 248 odd stipple */
            { ZXgrey(248, 7); return; }
         break;
      case ',':
         if (ch2 == ',')  /* 248 odd stipple */
            { ZXgrey(7, 248); return; }
         break;
      case '@':
         if (ch2 == '@')  /* 246 even stipple */
            { ZXgrey(246, 246); return; }
         break;
      case '!':
         if (ch2 == '!')  /* 246 even stipple */
            { ZXgrey(246, 0); return; }
         break;
      case ';':
         if (ch2 == ';')  /* 246 even stipple */
            { ZXgrey(0, 246); return; }
         break;
      case ' ':
         if (ch2 == ':')
            { ZXblock(7, 0, 7, 0); return; }
         if (ch2 == '\'')
            { ZXblock(7, 0, 7, 7); return; }
         if (ch2 == '.')
            { ZXblock(7, 7, 7, 0); return; }
         if (ch2 == ' ')
            { ZXblock(7, 7, 7, 7); return; }
               return;
         break;
      case ':':
         if (ch2 == ':')
            { ZXblock(0, 0, 0, 0); return; }
         if (ch2 == '\'')
            { ZXblock(0, 0, 0, 7); return; }
         if (ch2 == '.')
            { ZXblock(0, 7, 0, 0); return; }
         if (ch2 == ' ')
            { ZXblock(0, 7, 0, 7); return; }
         break;
      case '\'':
         if (ch2 == ':')
            { ZXblock(0, 0, 7, 0); return; }
         if (ch2 == '\'')
            { ZXblock(0, 0, 7, 7); return; }
         if (ch2 == '.')
            { ZXblock(0, 7, 7, 0); return; }
         if (ch2 == ' ')
            { ZXblock(0, 7, 7, 7); return; }
         break;
      case '.':
         if (ch2 == ':')
            { ZXblock(7, 0, 0, 0); return; }
         if (ch2 == '\'')
            { ZXblock(7, 0, 0, 7); return; }
         if (ch2 == '.')
            { ZXblock(7, 7, 0, 0); return; }
         if (ch2 == ' ')
            { ZXblock(7, 7, 0, 7); return; }
         break;
   }

   /* Question mark for unknown characters */
   ZXputc('?', 0, 7);
}


/* create specfic ZX81 grey character */
ZXgrey (up, lo)
   int up, lo;
{
   block(stdout, 8, 4, col, row, up);
   block(stdout, 8, 4, col, row+4, lo);
}


/* create ZX81 block characters */
ZXblock (ul, ur, ll, lr)
   int ul, ur, ll, lr;
{
   block(stdout, 4, 4, col, row, ul);
   block(stdout, 4, 4, col+4, row, ur);
   block(stdout, 4, 4, col, row+4, ll);
   block(stdout, 4, 4, col+4, row+4, lr);
}


/* move cursor down */
ZXdowncurs ()
{
   col = FIRSTCOL;
   row = row + 8;
}


/* adjust cursor for end of line/screen */
ZXadjustcurs ()
{
   col = col + 8;
   if (col > LASTCOL)
   {
      col = FIRSTCOL;
      row = row + 8;
   }
}
